/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2009-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */


/*!
    \page lcmaps_c_pep.mod C PEP for Argus plugin

    \section posixsynopsis SYNOPSIS
    \b lcmaps_c_pep.mod

    \section posixdesc DESCRIPTION

The GUMS module will connect to the GUMS server using the SAML protocol (provided by the Privilege Project) and return a Unix UID, GID and possibly secondary GIDs to the framework.


*/


/*!
    \file   lcmaps_c_pep.c
    \brief  Plugin to connect to a PEPd and return a UID, a GID and secondary GIDs.
    \author Oscar Koeroo for the EGEE.
*/





/*****************************************************************************
                            Include header files
******************************************************************************/
#include "lcmaps_config.h"

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <pwd.h>
#include <grp.h>
#include <time.h>

#include <netdb.h>
#include <sys/utsname.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <limits.h>
#include <fcntl.h>

#include <lcmaps/lcmaps_modules.h>
#include <lcmaps/lcmaps_arguments.h>
#include <lcmaps/lcmaps_cred_data.h>
#include <lcmaps/lcmaps_vo_data.h>

#include <lcmaps/lcmaps_account.h>

#include "lcmaps_c_pep.h"
#include "pep-c-obligation-handlers_helpers.h"


/******************************************************************************
                                Definitions
******************************************************************************/

#define PLUGIN_RUN      0
#define PLUGIN_VERIFY   1


// What do we want to do with this time presentation?
#define GMT               1
#define LOCAL             2
#define LEAVE_TIME_ITS_OK 3
#define LCMAPS_MAX_URNURL_LENGTH 512

/******************************************************************************
                          Module specific prototypes
******************************************************************************/

int plugin_initialize (int, char **);
int plugin_introspect(int *, lcmaps_argument_t **);
int plugin_run(int, lcmaps_argument_t *);
int plugin_verify(int, lcmaps_argument_t *);
int plugin_terminate(void);

static int plugin_run_or_verify(int, lcmaps_argument_t *, int);

char *plugin_c_pep_readpwfile(char *fn, char *buf, size_t n);
int plugin_c_pep_optequal(const char *s, const char *t);
char *plugin_c_pep_set_capath_default(char **p);
char *lcmaps_plugin_c_pep_cleanse_pemstring(char *in, char **destp);

int downgradeEffectiveToRealUid (uid_t * real_uid, uid_t * saved_uid);
int upgradeEffectiveToRealUid (uid_t real_uid, uid_t saved_uid);

char * readProxyFile (const char * proxyfile); 

/******************************************************************************
                       Define module specific variables
******************************************************************************/

#define MAX_ENDPOINTS         32
#define MAX_ENDPOINT_BUFSIZE  256
#define MAX_PWLEN             2048

#define PEP_OPTION_SSL_VALIDATION_DEFAULT -1
#define PEP_OPTION_SSL_VALIDATION_OFF 0
#define PEP_OPTION_SSL_VALIDATION_ON 1
#define PEP_OPTION_SSL_VALIDATION_ON 1

#define PEP_OPTION_CLIENT_CERT_UNDEFINED 0
#define PEP_OPTION_CLIENT_CERT_IMPLICIT  1
#define PEP_OPTION_CLIENT_CERT_EXPLICIT  2
#define PEP_OPTION_CLIENT_CERT_NONE      3

/* #ifndef XACML_GRIDWN_PROFILE_VERSION */
    /* #define XACML_GRIDWN_PROFILE_VERSION "http://glite.org/xacml/profile/grid-wn/1.0" */
/* #endif */
/* #ifndef XACML_AUTHZ_INTEROP_PROFILE_VERSION */
    /* #define XACML_INTEROP_PROFILE_VERSION "http://authz-interop.org/profile/1.1" */
/* #endif */
/*  */
/* #define PROFILE_AUTHZ_INTEROP 1 */
/* #define PROFILE_SWITCH_GRIDWN 2 */

static int    use_profile = PROFILE_SWITCH_GRIDWN; /* Use the new Grid WN profile from Switch as default */
    
static char   pep_daemon_endpoint_urls[MAX_ENDPOINTS][MAX_ENDPOINT_BUFSIZE];
static int    n_endpoints = 0;
static int    endpoint_selection_strategy = ENDPOINT_SELECTION_ROUND_ROBIN_TOP_TO_BOTTOM;
static int    pep_option_ssl_validation = PEP_OPTION_SSL_VALIDATION_DEFAULT;
static char * pep_option_server_cert= NULL;
static char * pep_option_server_capath = NULL;
static char * pep_option_server_capath_default = NULL;
static char * pep_option_client_cert= NULL;
static char * pep_option_client_key= NULL;
static char * pep_option_client_keypassword= NULL;
static char * pep_option_client_keypasswordfile= NULL;
static int    pep_option_client_cert_mode = PEP_OPTION_CLIENT_CERT_UNDEFINED;
static char   keypasswordbuffer[MAX_PWLEN];


static char * resourcetype            = NULL;
static char * actiontype              = NULL;
static char * resourceid              = NULL;
static char * actionid                = NULL;

static int    pep_c_debug             = 0;


/******************************************************************************
Function:   plugin_initialize
Description:
    Initialize plugin
Parameters:
    argc, argv
    argv[0]: the name of the plugin
Returns:
    LCMAPS_MOD_SUCCESS : succes
    LCMAPS_MOD_FAIL    : failure
    LCMAPS_MOD_NOFILE  : db file not found (will halt LCMAPS initialization)
******************************************************************************/
int plugin_initialize(int argc, char **argv)
{
    char * logstr = "\tlcmaps_plugin_c_pep-plugin_initialize()";
    int    i = 0;

    lcmaps_log_debug(2,"%s: passed arguments:\n", logstr);
    for (i=0; i < argc; i++)
    {
        lcmaps_log_debug(2,"%s: arg %d is %s\n", logstr, i, argv[i]);
    }

    /*
     * Parse arguments, argv[0] = name of plugin, so start with i = 1
     */
    for (i = 1; i < argc; i++)
    {
        /* Endpoint selection strategy */
	if ( plugin_c_pep_optequal(argv[i],"endpoint-strategy") 
	     && (i + 1 < argc) )
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                if      (strcasecmp (argv[i + 1], "round-robin") == 0)              endpoint_selection_strategy = ENDPOINT_SELECTION_ROUND_ROBIN_TOP_TO_BOTTOM;
                else if (strcasecmp (argv[i + 1], "round-robin-random-start") == 0) endpoint_selection_strategy = ENDPOINT_SELECTION_ROUND_ROBIN_RANDOM_START;
                else if (strcasecmp (argv[i + 1], "random") == 0)                   endpoint_selection_strategy = ENDPOINT_SELECTION_RANDOM;
                else
                {
                    lcmaps_log (1, "%s: Error: --endpoint-strategy not understood. Options are: \"round-robin\", \"round-robin-random-start\" and \"random\".\n", logstr);
                    return LCMAPS_MOD_FAIL;
                }
            }
            i++;
        }

        /* endpoints */
        else if ( plugin_c_pep_optequal(argv[i], "pep-daemon-endpoint-url") 
             && (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                if (n_endpoints < MAX_ENDPOINTS)
                {
                    bzero (pep_daemon_endpoint_urls[n_endpoints], MAX_ENDPOINT_BUFSIZE);
                    if (strlen (argv[1 + 1]) >= MAX_ENDPOINT_BUFSIZE)
                    {
                        lcmaps_log (1, "%s: Error: endpoint URL is too long: %s...\n", strncpy(pep_daemon_endpoint_urls[n_endpoints], argv[i + 1], MAX_ENDPOINT_BUFSIZE - 1), logstr);
                        return LCMAPS_MOD_FAIL;
                    }
                    else
                    {
                        strncpy(pep_daemon_endpoint_urls[n_endpoints], argv[i + 1], MAX_ENDPOINT_BUFSIZE - 1);
                        n_endpoints++;
                    }
                }
                else
                {
                    lcmaps_log (1, "%s: Error: Too many endpoints configured, the maximum amount of configurable endpoints is %d\n", logstr, MAX_ENDPOINTS);
                    return LCMAPS_MOD_FAIL;
                }
            }
            i++;
        }

        /* Resource Type - Set the resource type to CE, WN, RB, SE */
        else if ( plugin_c_pep_optequal(argv[i], "resourcetype")
                  && (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                char * tmprt = argv[i + 1];

                if (strcasecmp (tmprt, "se") == 0)      resourcetype = XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_SE;
                else if (strcasecmp (tmprt, "ce") == 0) resourcetype = XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_CE;
                else if (strcasecmp (tmprt, "wn") == 0) resourcetype = XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_WN;
                else if (strcasecmp (tmprt, "rb") == 0) resourcetype = XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_RB;
                else
                {
                    lcmaps_log (1, "%s: Error: Resource type unknown: %s\n", logstr, tmprt);
                    return LCMAPS_MOD_FAIL;
                }
            }
            i++;
        }

        /* Action Type - Set the action type to CE, WN, RB, SE */
        else if ( plugin_c_pep_optequal(argv[i], "actiontype")
                  && (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                char * tmprt = argv[i + 1];

                if (strcasecmp (tmprt, "access-file") == 0)      actiontype = XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_TYPE_ACCESS_FILE;
                else if (strcasecmp (tmprt, "queue") == 0)       actiontype = XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_TYPE_QUEUE;
                else if (strcasecmp (tmprt, "execute-now") == 0) actiontype = XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_TYPE_EXEC_NOW;
                else
                {
                    lcmaps_log (1, "%s: Error: Action type unknown: %s\n", logstr, tmprt);
                    return LCMAPS_MOD_FAIL;
                }
            }
            i++;
        }

        /* Resource ID - Override of the resource id. XACML spec dictates this to be changeable */
        else if ( plugin_c_pep_optequal(argv[i], "resourceid")
                  && (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                if (strlen(argv[i + 1]) > LCMAPS_MAX_URNURL_LENGTH)
                {
                    lcmaps_log (1, "%s: Error: Resource ID URL or URN to long.\n", logstr);
                    return LCMAPS_MOD_FAIL;
                }
                else
                    resourceid = argv[i + 1];
            }
            i++;
        }

        /* Action ID - Override of the action id. XACML spec dictates this to be changeable */
        else if ( plugin_c_pep_optequal(argv[i], "actionid") && (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                if (strlen(argv[i + 1]) > LCMAPS_MAX_URNURL_LENGTH)
                {
                    lcmaps_log (1, "%s: Error: Resource ID URL or URN to long.\n", logstr);
                    return LCMAPS_MOD_FAIL;
                }
                else
                    actionid = argv[i + 1];
            }
            i++;
        }

        /* Debug setting for the PEP-C library (will log on stderr) */
        else if ( plugin_c_pep_optequal(argv[i], "pep-c-debug") )
        {
            pep_c_debug = 1;
        }

        /* force the PEP-C library to do SSL validation checks, regardless of its default setting */
        else if ( plugin_c_pep_optequal(argv[i],  "check-certificates") )
        {
            pep_option_ssl_validation = PEP_OPTION_SSL_VALIDATION_ON;
        }

        /* force the PEP-C library to ignore SSL validation checks */
        else if ( plugin_c_pep_optequal(argv[i], "no-check-certificates") )
        {
            pep_option_ssl_validation = PEP_OPTION_SSL_VALIDATION_OFF;
        }

	/* CApath */
        else if ( ( plugin_c_pep_optequal(argv[i], "capath") || plugin_c_pep_optequal(argv[i], "CApath") )
                  && (i + 1 < argc) )
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                if (strlen(argv[i + 1]) > PATH_MAX)
                {
                    lcmaps_log (1, "%s: Error: CApath too long in argument.\n", logstr);
                    return LCMAPS_MOD_FAIL;
                }
                else
                    pep_option_server_capath = argv[i + 1];
            }
            i++;
        }
        /* file with a CA certificate or set of certs */
        else if ( plugin_c_pep_optequal(argv[i], "cafile") && (i + 1 < argc) )
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                if (strlen(argv[i + 1]) > PATH_MAX)
                {
                    lcmaps_log (1, "%s: Error: ca file name too long in argument.\n", logstr);
                    return LCMAPS_MOD_FAIL;
                }
                else
                    pep_option_server_cert = argv[i + 1];
            }
            i++;
        }

        /* use implicit user proxy, explicit file or no client cert at all */
        else if ( plugin_c_pep_optequal(argv[i], "pep-certificate-mode") && (i + 1 < argc) )
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                char * tmprt = argv[i + 1];

                if (strcasecmp (tmprt, "implicit") == 0)      pep_option_client_cert_mode = PEP_OPTION_CLIENT_CERT_IMPLICIT;
                else if (strcasecmp (tmprt, "explicit") == 0) pep_option_client_cert_mode = PEP_OPTION_CLIENT_CERT_EXPLICIT;
                else if (strcasecmp (tmprt, "none") == 0)     pep_option_client_cert_mode = PEP_OPTION_CLIENT_CERT_NONE;
                else
                {
                    lcmaps_log (1, "%s: Error: pep certificate mode unknown: %s (should be implicit, explicit, or none)\n", logstr, tmprt);
                    return LCMAPS_MOD_FAIL;
                }
            }
            i++;
        }
	/* certificate to use to contact PEPd (other curl SSL settings) */
        else if ( plugin_c_pep_optequal(argv[i], "certificate") && (i + 1 < argc) )
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                if (strlen(argv[i + 1]) > PATH_MAX)
                {
                    lcmaps_log (1, "%s: Error: certificate file name too long in argument.\n", logstr);
                    return LCMAPS_MOD_FAIL;
                }
                else
                    pep_option_client_cert = argv[i + 1];
            }
            i++;
        }

	/* certificate to use to contact PEPd (other curl SSL settings) */
        else if ( plugin_c_pep_optequal(argv[i], "key") && (i + 1 < argc) )
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                if (strlen(argv[i + 1]) > PATH_MAX)
                {
                    lcmaps_log (1, "%s: Error: certificate key name too long in argument.\n", logstr);
                    return LCMAPS_MOD_FAIL;
                }
                else
                    pep_option_client_key = argv[i + 1];
            }
            i++;
        }

	/* certificate to use to contact PEPd (other curl SSL settings) */
        else if ( plugin_c_pep_optequal(argv[i], "passfile") && (i + 1 < argc) )
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                if (strlen(argv[i + 1]) > PATH_MAX)
                {
                    lcmaps_log (1, "%s: Error: passwork file name too long in argument.\n", logstr);
                    return LCMAPS_MOD_FAIL;
                }
                else
                    pep_option_client_keypasswordfile = argv[i + 1];
            }
            i++;
        }

	/* certificate to use to contact PEPd (other curl SSL settings) */
        else if ( plugin_c_pep_optequal(argv[i], "pass") && (i + 1 < argc) )
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                if (strlen(argv[i + 1]) > MAX_PWLEN)
                {
                    lcmaps_log (1, "%s: Error: password too long in argument.\n", logstr);
                    return LCMAPS_MOD_FAIL;
                }
                else
                    pep_option_client_keypassword = argv[i + 1];
            }
            i++;
        }

	    /* Profile choice */
        else if ( plugin_c_pep_optequal(argv[i], "profile") && (i + 1 < argc) )
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                if (strncasecmp(argv[i + 1], XACML_GRIDWN_PROFILE_VERSION, strlen(XACML_GRIDWN_PROFILE_VERSION)) == 0)
                {
                    use_profile = PROFILE_SWITCH_GRIDWN;
                }
                else if (strncasecmp(argv[i + 1], XACML_INTEROP_PROFILE_VERSION, strlen(XACML_INTEROP_PROFILE_VERSION)) == 0)
                {
                    use_profile = PROFILE_AUTHZ_INTEROP;
                }
                else
                {
                    lcmaps_log (1, "%s: Error: the configured profile is not supported.\n", logstr);
                    lcmaps_log (1, "%s:        Supported profiles are:\n", logstr);
                    lcmaps_log (1, "%s:        - %s (default)\n", logstr, XACML_GRIDWN_PROFILE_VERSION);
                    lcmaps_log (1, "%s:        - %s\n",           logstr, XACML_INTEROP_PROFILE_VERSION);
                    return LCMAPS_MOD_FAIL;
                }
            }
            i++;
        }

        /* Parameter unknown */
        else
        {
            lcmaps_log(1,"%s: Error in initialization parameter: %s (failure)\n", logstr, argv[i]);
            return LCMAPS_MOD_FAIL;
        }
    }

    /* If no cert-mode was specified, but an explicit cert was given, assume this was intended */
    if ( ( PEP_OPTION_CLIENT_CERT_UNDEFINED == pep_option_client_cert_mode ) 
         && ( pep_option_client_cert || pep_option_client_key ) ) {
      pep_option_client_cert_mode = PEP_OPTION_CLIENT_CERT_EXPLICIT;
    } 
    /* Otherwise, the default is 'none' to make it work without config for glexec-on-WNs, but
     * in that caase you will not have client auth (anyone can impersone anybody). If you want
     * client auath in the glexec-on-WN scenario, you need the implicit mode, but this may break
     * when the proxy is on a NFS root-squashed file system. But fixing that bug is disallowed
     * by policy. */
    if ( PEP_OPTION_CLIENT_CERT_UNDEFINED == pep_option_client_cert_mode )  {
      pep_option_client_cert_mode = PEP_OPTION_CLIENT_CERT_NONE;
    }

    /* Basic sanity checks on the certificate file specification that only makes sense in explicit mode */
    if ( pep_option_client_cert && ( PEP_OPTION_CLIENT_CERT_EXPLICIT != pep_option_client_cert_mode ) ) {
        lcmaps_log (1, "%s: Error: certificate file must be used in explicit mode only.\n", logstr);
        return LCMAPS_MOD_FAIL;
    }
    if ( ( PEP_OPTION_CLIENT_CERT_EXPLICIT == pep_option_client_cert_mode ) && 
         ( (!pep_option_client_cert) || (!pep_option_client_key) ) ) {
        lcmaps_log (1, "%s: Error: explicit certificate mode requires a certificate file and a key.\n", logstr);
        return LCMAPS_MOD_FAIL;
    }
    if ( ( PEP_OPTION_CLIENT_CERT_IMPLICIT == pep_option_client_cert_mode ) && !getenv("X509_USER_PROXY") ) {
        lcmaps_log (1, "%s: Error: PEPd communication in implicit mode requires proxy filename in X509_USER_PROXY.\n", logstr);
        return LCMAPS_MOD_FAIL;
    }
    lcmaps_log_debug (5, "%s: PEPd client certificate mode set to %d.\n", logstr, pep_option_client_cert_mode);


    /* password should either be in a file OR in the command line, not both */
    if ( pep_option_client_keypassword && pep_option_client_keypasswordfile ) {
        lcmaps_log (1, "%s: Error: PEPd client key password should either be in a file OR in the command line, not both.\n", logstr);
        return LCMAPS_MOD_FAIL;
    }
    if ( pep_option_client_keypasswordfile ) {
      pep_option_client_keypassword = plugin_c_pep_readpwfile(pep_option_client_keypasswordfile,keypasswordbuffer,MAX_PWLEN);
      if ( !pep_option_client_keypassword ) {
        lcmaps_log (1, "%s: Error: PEPd client key password could not be read from file %s.\n", logstr, pep_option_client_keypasswordfile);
        return LCMAPS_MOD_FAIL;
      }
    }

    /* the CApath is taken from the command line argument in pep_option_server_capath
     * if set, but otherwise defaults 
     */
    if ( !pep_option_server_capath ) {
      pep_option_server_capath = plugin_c_pep_set_capath_default(&pep_option_server_capath_default);
    }


    switch (endpoint_selection_strategy)
    {
        case ENDPOINT_SELECTION_ROUND_ROBIN_TOP_TO_BOTTOM :
            lcmaps_log_debug (5, "%s: Endpoint selection strategy set to \"round-robin, starting from the top of the list\"\n", logstr);
            break;
        case ENDPOINT_SELECTION_ROUND_ROBIN_RANDOM_START  :
            lcmaps_log_debug (5, "%s: Endpoint selection strategy set to \"round-robin with random starting point in the list\"\n", logstr);
            break;
        case ENDPOINT_SELECTION_RANDOM                    :
            lcmaps_log_debug (5, "%s: Endpoint selection strategy set to \"random endpoint selection\"\n", logstr);
            break;
    }


    lcmaps_log_debug(2, "%s: Initialization succeeded\n", logstr);

    return LCMAPS_MOD_SUCCESS;
}


/******************************************************************************
Function:   plugin_introspect
Description:
    return list of required arguments
Parameters:

Returns:
    LCMAPS_MOD_SUCCESS : succes
    LCMAPS_MOD_FAIL    : failure
******************************************************************************/
int plugin_introspect(int *argc, lcmaps_argument_t **argv)
{
  char *logstr = "\tlcmaps_plugin_c_pep-plugin_introspect()";
  static lcmaps_argument_t argList[] = {
    {"pem_string", "char *"  ,  0, NULL},
    {"user_dn"   , "char *"  ,  0, NULL},
    {NULL        ,  NULL     , -1, NULL}
  };

  lcmaps_log_debug(3, "%s: introspecting\n", logstr);

  *argv = argList;
  *argc = lcmaps_cntArgs(argList);
  lcmaps_log_debug(5, "%s: address first argument: 0x%x\n", logstr, argList);

  lcmaps_log_debug(2, "%s: Introspect succeeded\n", logstr);

  return LCMAPS_MOD_SUCCESS;
}




/******************************************************************************
Function:   plugin_run
Description:
    Gather credentials for LCMAPS
Parameters:
    argc: number of arguments
    argv: list of arguments
Returns:
    LCMAPS_MOD_SUCCESS: authorization succeeded
    LCMAPS_MOD_FAIL   : authorization failed
******************************************************************************/
int plugin_run(int argc, lcmaps_argument_t * argv)
{
    return plugin_run_or_verify(argc, argv, PLUGIN_RUN);
}


/******************************************************************************
Function:   plugin_verify
Description:
    Verify if user is entitled to use local credentials based on his grid
    credentials. This means that the site should already have been set up
    by, e.g., LCMAPS in a previous run. This method will not try to setup
    account leases, modify (distributed) passwd/group files, etc. etc.
    The outcome should be identical to that of plugin_run().
    In this particular case "plugin_verify()" is identical to "plugin_run()"

Parameters:
    argc: number of arguments
    argv: list of arguments
Returns:
    LCMAPS_MOD_SUCCESS: authorization succeeded
    LCMAPS_MOD_FAIL   : authorization failed
******************************************************************************/
int plugin_verify(int argc, lcmaps_argument_t * argv)
{
  return plugin_run_or_verify(argc, argv, PLUGIN_VERIFY);
}

static int plugin_run_or_verify(int argc, lcmaps_argument_t *argv, int lcmaps_mode)
{
    char *logstr = NULL;
    char *datetime = NULL;
    time_t my_clock;
    struct tm *tmpTime;
    /* char *encoded_dn = "(void)"; */
    char * dnshostname = NULL;
    int pepd_success = 0;
    char * dn = NULL;
    int    dn_cnt = 0;

    int                 i = 0;
#if 0
    unsigned int        seed;    
    int                 random_choice = 0;
#endif
    int rc = 0;

    char *pem_string = NULL;
    char *clean_pem_string = NULL;

    char *proxy_file_pilot = NULL;
    char *pem_string_pilot = NULL;
    char *clean_pem_string_pilot = NULL;

    uid_t stored_real_uid = -1;
    uid_t stored_eff_uid  = -1;

#if PEP_API != 1
    PEP *pep_handle=NULL;
#endif

    time(&my_clock);
    tmpTime = gmtime(&my_clock);
    datetime = malloc(sizeof(char) * 20);
    snprintf(datetime, 20, "%04d-%02d-%02d %02d:%02d:%02d",
             tmpTime->tm_year + 1900, tmpTime->tm_mon + 1, tmpTime->tm_mday,
             tmpTime->tm_hour, tmpTime->tm_min, tmpTime->tm_sec);

    if (lcmaps_mode == PLUGIN_RUN)
        logstr = "lcmaps_plugin_c_pep-plugin_run()";
    else if (lcmaps_mode == PLUGIN_VERIFY)
        logstr = "lcmaps_plugin_c_pep-plugin_verify()";
    else {
        lcmaps_log(1, "\tlcmaps_plugin_c_pep-plugin_run_or_verify(): attempt to run plugin in invalid mode: %d\n", lcmaps_mode);

        goto c_pep_failure;
        /* return(lcmaps_plugin_c_pep_failed(logstr, datetime)); */
    }

    /* Grabbing DN */
    if ( ( dn = *(char **) lcmaps_getArgValue("user_dn", "char *", argc, argv) ) )
    {
        lcmaps_log_debug(1,"%s: found dn: %s\n", logstr, dn);

        /* Do I already have a DN registered in LCMAPS? If not registered, push it to the end-result registry */
        getCredentialData (DN, &dn_cnt);
        if (dn_cnt == 0)
        {
            lcmaps_log_debug(5,"%s: registering DN.\n", logstr, dn);
            addCredentialData(DN, &dn);
        }
    }
    else
        lcmaps_log(1,"%s: could not get value of dn !\n", logstr);



    /* Grab PEM string when available */
    if ((pem_string = *(char **) lcmaps_getArgValue("pem_string", "char *", argc, argv))) 
    {
        lcmaps_log_debug(2, "%s: found PEM string\n", logstr);
    } 
    else 
    {
        lcmaps_log(1, "%s: could not get value of PEM string from LCMAPS framework.\n", logstr);
        goto c_pep_failure;
    }

    /* the PEM string may contain a private key, which is used in verify_proxy and
     * friends. The P*P AuthZ framework, however, does not need the private key. Remove
     * it here to prevent needless exposure of the key. Don't mangle the incoming string
     * since it may be a static copy from the framework structure.
     */
    pem_string = lcmaps_plugin_c_pep_cleanse_pemstring(pem_string,&clean_pem_string);
    if ( !pem_string ) {
        lcmaps_log(1, "%s: could not remove the private key from the read proxyfile (of the user)\n", logstr);
        goto c_pep_failure;
    }

    /* Do work here */

    lcmaps_log_debug (5, "%s: Fetching DNS/Host name\n", logstr);
  #ifndef MAXHOSTNAMELEN
       #define MAXHOSTNAMELEN 255
  #endif /* MAXHOSTNAMELEN */
    dnshostname = (char *) malloc (sizeof(char) * MAXHOSTNAMELEN);
    gethostname (dnshostname, MAXHOSTNAMELEN);
    lcmaps_log_debug (5, "%s: DNS/Host name is: %s\n", logstr, dnshostname);
    

    lcmaps_log_debug (5, "PEP-C: initialize()\n");
#if PEP_API == 1
    if (pepc_initialize (pep_c_debug))
#else
    if ((pep_handle = pepc_initialize (pep_c_debug))==NULL)
#endif
    {
        lcmaps_log (1, "Error: Failed to initialize the PEP-C library\n");
        goto c_pep_failure;
    }

    /* Fetch the pilot's proxy file (if set) to be put into the SAML2-XACML2 
     * environment and used in the client side authentication for the SSL
     */
    proxy_file_pilot = getenv("X509_USER_PROXY");
    pem_string_pilot = readProxyFile (proxy_file_pilot);


    /* the PEM string may contain a private key, which is used in verify_proxy and
     * friends. The P*P AuthZ framework, however, does not need the private key. Remove
     * it here to prevent needless exposure of the key. Don't mangle the incoming string
     * since it may be a static copy from the framework structure.
     */
    if (!(lcmaps_plugin_c_pep_cleanse_pemstring(pem_string_pilot, &clean_pem_string_pilot)))
    {
        lcmaps_log(1, "%s: could not remove the private key from the read proxyfile (of the Pilot)\n", logstr);
        goto c_pep_failure;
    }

    lcmaps_log_debug (5, "PEP-C: construct_request()\n");
    if (pepc_construct_request (
#if PEP_API != 1
				pep_handle,
#endif
				use_profile,
                                pem_string, 
                                actiontype,
                                actionid,
                                resourceid,
                                resourcetype,
                                dnshostname,
                                clean_pem_string_pilot))
    {
        lcmaps_log (1, "Error: Failed to construct the Request\n");
        goto c_pep_failure;
    }


#if PEP_API == 1
    /* Load all the PEP daemon endpoints to the library */
    for (i = 0; i < n_endpoints; i++)
    {
        lcmaps_log_debug (2, "%s: C-PEP: Selecting PEP Daemon URL: \"%s\"\n",
			  logstr, pep_daemon_endpoint_urls[i]);
        rc = pep_setoption(PEP_OPTION_ENDPOINT_URL, pep_daemon_endpoint_urls[i]);
        if (rc != PEP_OK) {
            lcmaps_log (1, "Error: pep_setoption(PEP_OPTION_ENDPOINT_URL,%s) failed: %d: %s\n",
			pep_daemon_endpoint_urls[i], rc, pep_strerror(rc));
            goto c_pep_failure;
        }
    }

    /* Set pep options */

    /* PEP_OPTION_ENDPOINT_SSL_VALIDATION */
    if ( pep_option_ssl_validation != PEP_OPTION_SSL_VALIDATION_DEFAULT ) {
        if ( pep_option_ssl_validation == PEP_OPTION_SSL_VALIDATION_ON ) {
            rc = pep_setoption(PEP_OPTION_ENDPOINT_SSL_VALIDATION, 1);
            if (rc != PEP_OK) {
                lcmaps_log (1, "Error: pep_setoption(PEP_OPTION_ENDPOINT_SSL_VALIDATION,1) failed: %d: %s\n", 
                            rc, pep_strerror(rc));
                goto c_pep_failure;
            }
        } else if ( pep_option_ssl_validation == PEP_OPTION_SSL_VALIDATION_OFF ) {
            rc = pep_setoption(PEP_OPTION_ENDPOINT_SSL_VALIDATION, 0);
            if (rc != PEP_OK) {
                lcmaps_log (1, "Error: pep_setoption(PEP_OPTION_ENDPOINT_SSL_VALIDATION,0) failed: %d: %s\n", 
                            rc, pep_strerror(rc));
                goto c_pep_failure;
            }
        } else {
            lcmaps_log (1, "LCMAPS internal error: pep_option_ssl_validation set to invalid value (%d)\n",
                        pep_option_ssl_validation);
            goto c_pep_failure;
        }
    }

    /* PEP_OPTION_ENDPOINT_CLIENT_CERT and PEP_OPTION_ENDPOINT_CLIENT_KEY */
    if ( PEP_OPTION_CLIENT_CERT_EXPLICIT == pep_option_client_cert_mode ) {
        if ( !pep_option_client_cert || !pep_option_client_key ) {
            lcmaps_log (1, "Error: assert failed: setoptions for explicit client cert and key needs both set");
            goto c_pep_failure;
        }
        rc = pep_setoption(PEP_OPTION_ENDPOINT_CLIENT_CERT, pep_option_client_cert);
        if (rc != PEP_OK) {
            lcmaps_log (1, "Error: pep_setoption(PEP_OPTION_ENDPOINT_CLIENT_CERT,%s) failed: %d: %s\n", 
                        pep_option_client_cert, rc, pep_strerror(rc));
            goto c_pep_failure;
        }
        rc = pep_setoption(PEP_OPTION_ENDPOINT_CLIENT_KEY, pep_option_client_key);
        if (rc != PEP_OK) {
            lcmaps_log (1, "Error: pep_setoption(PEP_OPTION_ENDPOINT_CLIENT_KEY,%s) failed: %d: %s\n", 
                        pep_option_client_key, rc, pep_strerror(rc));
            goto c_pep_failure;
        }
    } else if ( PEP_OPTION_CLIENT_CERT_IMPLICIT == pep_option_client_cert_mode ) {
        if ( !proxy_file_pilot ) {
            lcmaps_log (1, "Error: assert failed: setoptions for implicit ssl needs X509_USER_PROXY set");
            goto c_pep_failure;
        }
        /* This is not the case anymore as the PEPd will change it behavior in a future release of Argus */
        /* lcmaps_log (1, "Warning: implicit SSL client auth with invokers proxy will not influence authorization decision by PEPd"); */
        rc = pep_setoption(PEP_OPTION_ENDPOINT_CLIENT_CERT, proxy_file_pilot);
        if (rc != PEP_OK) {
            lcmaps_log (1, "Error: pep_setoption(PEP_OPTION_ENDPOINT_CLIENT_CERT,%s) failed: %d: %s\n", 
                        proxy_file_pilot, rc, pep_strerror(rc));
            goto c_pep_failure;
        }
        rc = pep_setoption(PEP_OPTION_ENDPOINT_CLIENT_KEY, proxy_file_pilot);
        if (rc != PEP_OK) {
            lcmaps_log (1, "Error: pep_setoption(PEP_OPTION_ENDPOINT_CLIENT_KEY,%s) failed: %d: %s\n", 
                        proxy_file_pilot, rc, pep_strerror(rc));
            goto c_pep_failure;
        }
    } else if ( PEP_OPTION_CLIENT_CERT_NONE == pep_option_client_cert_mode ) {
        lcmaps_log_debug (5, "pep_option_client_cert_mode set to none\n");
    } else {
        lcmaps_log (1, "LCMAPS internal error: pep_option_client_cert_mode set to invalid value  at this point (%d)\n",
                    pep_option_ssl_validation);
        goto c_pep_failure;
    }

    /* PEP_OPTION_ENDPOINT_CLIENT_KEYPASSWORD */
    if ( pep_option_client_keypassword ) {
        rc = pep_setoption(PEP_OPTION_ENDPOINT_CLIENT_KEYPASSWORD, pep_option_client_keypassword);
        if (rc != PEP_OK) {
            lcmaps_log (1, "Error: pep_setoption(PEP_OPTION_ENDPOINT_CLIENT_KEYPASSWORD,****SUPPRESSED****) failed: %d: %s\n", 
                        rc, pep_strerror(rc));
            goto c_pep_failure;
        }
    }

    /* PEP_OPTION_ENDPOINT_SERVER_CAPATH */
    if ( pep_option_server_capath ) {
        rc = pep_setoption(PEP_OPTION_ENDPOINT_SERVER_CAPATH, pep_option_server_capath);
        if (rc != PEP_OK) {
            lcmaps_log (1, "Error: pep_setoption(PEP_OPTION_ENDPOINT_SERVER_CAPATH,%s) failed: %d: %s\n", 
                        pep_option_server_capath, rc, pep_strerror(rc));
            goto c_pep_failure;
        }
    }

    /* PEP_OPTION_ENDPOINT_SERVER_CERT */
    if ( pep_option_server_cert ) {
        rc = pep_setoption(PEP_OPTION_ENDPOINT_SERVER_CERT, pep_option_server_cert);
        if (rc != PEP_OK) 
        {
            lcmaps_log (1, "Error: pep_setoption(PEP_OPTION_ENDPOINT_SERVER_CERT,%s) failed: %d: %s\n", 
                        pep_option_server_cert, rc, pep_strerror(rc));
            goto c_pep_failure;
        }
    }
/* PEP_API == 1 */
#else
/* PEP_API != 1 */
    /* Load all the PEP daemon endpoints to the library */
    for (i = 0; i < n_endpoints; i++)
    {
        lcmaps_log_debug (2, "%s: C-PEP: Selecting PEP Daemon URL: \"%s\"\n",
			  logstr, pep_daemon_endpoint_urls[i]);
        rc = pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_URL, pep_daemon_endpoint_urls[i]);
        if (rc != PEP_OK) {
            lcmaps_log (1, "Error: pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_URL,%s) failed: %d: %s\n",
			pep_daemon_endpoint_urls[i], rc, pep_strerror(rc));
            goto c_pep_failure;
        }
    }

    /* Set pep options */ 

    /* Set SSL cipher list, to be compatible with OpenSSL 1.0 client-side.
     * This setting is only available in pep-api 2.0. Setting it here is a
     * workaround, since this really should be fixed on server-side */
    rc=pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_SSL_CIPHER_LIST, "DEFAULT:-ECDH");
    if (rc != PEP_OK) {
	lcmaps_log (1, "Error: pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_SSL_CIPHER_LIST,\"DEFAULT:-ECDH\") failed: %d: %s\n", 
		    rc, pep_strerror(rc));
	goto c_pep_failure;
    }

    /* PEP_OPTION_ENDPOINT_SSL_VALIDATION */
    if ( pep_option_ssl_validation != PEP_OPTION_SSL_VALIDATION_DEFAULT ) {
        if ( pep_option_ssl_validation == PEP_OPTION_SSL_VALIDATION_ON ) {
            rc = pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_SSL_VALIDATION, 1);
            if (rc != PEP_OK) {
                lcmaps_log (1, "Error: pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_SSL_VALIDATION,1) failed: %d: %s\n", 
                            rc, pep_strerror(rc));
                goto c_pep_failure;
            }
        } else if ( pep_option_ssl_validation == PEP_OPTION_SSL_VALIDATION_OFF ) {
            rc = pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_SSL_VALIDATION, 0);
            if (rc != PEP_OK) {
                lcmaps_log (1, "Error: pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_SSL_VALIDATION,0) failed: %d: %s\n", 
                            rc, pep_strerror(rc));
                goto c_pep_failure;
            }
        } else {
            lcmaps_log (1, "LCMAPS internal error: pep_option_ssl_validation set to invalid value (%d)\n",
                        pep_option_ssl_validation);
            goto c_pep_failure;
        }

    }

    /* PEP_OPTION_ENDPOINT_CLIENT_CERT and PEP_OPTION_ENDPOINT_CLIENT_KEY */
    if ( PEP_OPTION_CLIENT_CERT_EXPLICIT == pep_option_client_cert_mode ) {
        if ( !pep_option_client_cert || !pep_option_client_key ) {
            lcmaps_log (1, "Error: assert failed: setoptions for explicit client cert and key needs both set");
            goto c_pep_failure;
        }
        rc = pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_CLIENT_CERT, pep_option_client_cert);
        if (rc != PEP_OK) {
            lcmaps_log (1, "Error: pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_CLIENT_CERT,%s) failed: %d: %s\n", 
                        pep_option_client_cert, rc, pep_strerror(rc));
            goto c_pep_failure;
        }
        rc = pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_CLIENT_KEY, pep_option_client_key);
        if (rc != PEP_OK) {
            lcmaps_log (1, "Error: pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_CLIENT_KEY,%s) failed: %d: %s\n", 
                        pep_option_client_key, rc, pep_strerror(rc));
            goto c_pep_failure;
        }
    } else if ( PEP_OPTION_CLIENT_CERT_IMPLICIT == pep_option_client_cert_mode ) {
        if ( !proxy_file_pilot ) {
            lcmaps_log (1, "Error: assert failed: setoptions for implicit ssl needs X509_USER_PROXY set");
            goto c_pep_failure;
        }
        /* This is not the case anymore as the PEPd will change it behavior in a future release of Argus */
        /* lcmaps_log (1, "Warning: implicit SSL client auth with invokers proxy will not influence authorization decision by PEPd"); */
        rc = pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_CLIENT_CERT, proxy_file_pilot);
        if (rc != PEP_OK) {
            lcmaps_log (1, "Error: pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_CLIENT_CERT,%s) failed: %d: %s\n", 
                        proxy_file_pilot, rc, pep_strerror(rc));
            goto c_pep_failure;
        }
        rc = pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_CLIENT_KEY, proxy_file_pilot);
        if (rc != PEP_OK) {
            lcmaps_log (1, "Error: pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_CLIENT_KEY,%s) failed: %d: %s\n", 
                        proxy_file_pilot, rc, pep_strerror(rc));
            goto c_pep_failure;
        }
    } else if ( PEP_OPTION_CLIENT_CERT_NONE == pep_option_client_cert_mode ) {
        lcmaps_log_debug (5, "pep_option_client_cert_mode set to none\n");
    } else {
        lcmaps_log (1, "LCMAPS internal error: pep_option_client_cert_mode set to invalid value  at this point (%d)\n",
                    pep_option_ssl_validation);
        goto c_pep_failure;
    }

    /* PEP_OPTION_ENDPOINT_CLIENT_KEYPASSWORD */
    if ( pep_option_client_keypassword ) {
        rc = pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_CLIENT_KEYPASSWORD, pep_option_client_keypassword);
        if (rc != PEP_OK) {
            lcmaps_log (1, "Error: pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_CLIENT_KEYPASSWORD,****SUPPRESSED****) failed: %d: %s\n", 
                        rc, pep_strerror(rc));
            goto c_pep_failure;
        }
    }

    /* PEP_OPTION_ENDPOINT_SERVER_CAPATH */
    if ( pep_option_server_capath ) {
        rc = pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_SERVER_CAPATH, pep_option_server_capath);
        if (rc != PEP_OK) {
            lcmaps_log (1, "Error: pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_SERVER_CAPATH,%s) failed: %d: %s\n", 
                        pep_option_server_capath, rc, pep_strerror(rc));
            goto c_pep_failure;
        }
    }

    /* PEP_OPTION_ENDPOINT_SERVER_CERT */
    if ( pep_option_server_cert ) {
        rc = pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_SERVER_CERT, pep_option_server_cert);
        if (rc != PEP_OK) {
            lcmaps_log (1, "Error: pep_setoption(pep_handle, PEP_OPTION_ENDPOINT_SERVER_CERT,%s) failed: %d: %s\n", 
                        pep_option_server_cert, rc, pep_strerror(rc));
            goto c_pep_failure;
        }
    }
#endif /* PEP_API != 2 */


    lcmaps_log_debug (5, "PEP-C: engage()\n");
#if PEP_API == 1
    if (pepc_engage ())
#else
    if (pepc_engage (pep_handle))
#endif
        lcmaps_log (2, "Error: An error has occured in the PEP Daemon interaction.\n");
    else
        pepd_success = 1;



    if (!pepd_success)
        goto c_pep_failure;


    lcmaps_log_debug (5, "PEP-C: cleanup()\n");

    /* Downgrade effective privileges to cope with NFS mounted file systems with root squashing */
    downgradeEffectiveToRealUid (&stored_real_uid, &stored_eff_uid);
#if PEP_API == 1
    if (pepc_cleanup ())
#else
    if (pepc_cleanup (pep_handle))
#endif
        lcmaps_log (1, "Error: Failed to clean up PEP-C (error subpressed)\n");

    /* Restore privileges to previous state */
    upgradeEffectiveToRealUid (stored_real_uid, stored_eff_uid);

    /*
     *  Add the pool index by hand. Normally the poolindex plug in takes
     *  care of this. Since the GUMS plugin setup the poolaccounts, we need
     *  to have a 'NULL' account inorder to prevent LCMAPS from complaining
     *  about missing poolaccounts.
     */
    /* lcmaps_log_debug (5, "%s: Adding POOL_INDEX :  %s\n", logstr, encoded_dn); */
    /* rc += addCredentialData(POOL_INDEX, &encoded_dn); */



    if (rc==0) {
       lcmaps_log_time(1, "%s: C-PEP plugin succeeded\n", logstr);
    } else {
       lcmaps_log (1, "%s Error: At least one of the addCredentialData calls failed.\n", logstr);
       goto c_pep_failure;
    }

    free (datetime);
    free (dnshostname);
    free (clean_pem_string);
    free (pem_string_pilot);
    free (clean_pem_string_pilot);

    /* Clean up supported obligations, not needed after registration */
    free_supported_obligations_list();

    return LCMAPS_MOD_SUCCESS;


c_pep_failure:
    lcmaps_log_debug (5, "PEP-C: cleanup()\n");
#if PEP_API == 1
    if (pepc_cleanup ())
#else
    if (pepc_cleanup (pep_handle))
#endif
        lcmaps_log (1, "Error: Failed to clean up PEP-C, but cannot bugger out anymore. Ignoring failure\n");

    if (datetime) free (datetime);
    if (dnshostname) free (dnshostname);
    if (clean_pem_string) free (clean_pem_string);
    if (pem_string_pilot) free (pem_string_pilot);
    if (clean_pem_string_pilot) free (clean_pem_string_pilot);
    lcmaps_log_time(1, "%s: c_pep plugin failed\n", logstr);

    return LCMAPS_MOD_FAIL;
}



/******************************************************************************
Function:   plugin_terminate
Description:
    Terminate plugin
Parameters:

Returns:
    LCMAPS_MOD_SUCCESS : succes
    LCMAPS_MOD_FAIL    : failure
******************************************************************************/
int plugin_terminate()
{
    char * logstr = "\tlcmaps_plugin_c_pep-plugin_terminate()";
    lcmaps_log_debug(2,"%s: terminating\n", logstr);

    /* return any memory allowed by subsidiaries */
    if ( pep_option_server_capath_default ) 
        free( pep_option_server_capath_default );

    return LCMAPS_MOD_SUCCESS;
}


/******************************************************************************
Function:   plugin_c_pep_readpwfile
Description:
    Read a password (single line) from a file
Parameters:
    static char *f, char *staticbuf, int n;
Returns:
    pointer to the buffer with the password or NULL on failure
******************************************************************************/

char *plugin_c_pep_readpwfile(char *fn, char *buf, size_t n)
{
  FILE *f;
  char *s;

  if ( !fn ) {
      lcmaps_log(1, "plugin_c_pep_readpwfile: called with an empty file name.\n");
      return NULL;
  }
  if ( ( n <= 0 ) || !buf ) {
      lcmaps_log(1, "plugin_c_pep_readpwfile: string buffer size zero or negative or buffer is a NULL pointer.\n");
      return NULL;
  }

  if ( !(f = fopen(fn,"r")) ) {
      lcmaps_log(1, "plugin_c_pep_readpwfile: cannot open password file %s.\n",fn);
      return NULL;
  }
  errno=0;
  s = fgets(buf,n,f);
  if ( !s ) {
    if ( errno ) { /* a true error condition */
      lcmaps_log(1, "plugin_c_pep_readpwfile: cannot read password file %s: error %d.\n", fn,errno);
      fclose(f);
      return NULL;
    } else {
      /* EOF reached, password is empty file */
      *buf = '\0';
    }
  }
  fclose(f);

  if ( strlen(buf) == n ) {
      lcmaps_log(1, "password in file %s is too long (at or over %d characters.\n", fn,n-1);
      return NULL;
  }

  /* strip new-line or CR characters for any platform */
  s = buf+strlen(buf);
  while ( ( s > buf ) && ('\n' == *s || '\r' == *s) ) { *s = '\0'; s--; }

  return buf;
}


/******************************************************************************
Function:   plugin_c_pep_optequal(s,t)
Description:
    See if <s> is the option "-[-]<t>"
Parameters:
    const char *s, const char *t
Returns:
    0	if <s> is NOT this option
    1	if <s> IS this option
    -1	on error
******************************************************************************/
int plugin_c_pep_optequal(const char *s, const char *t) 
{
	register int n,m;
	const char *p = s;

	if ( !s || !t ) {
		lcmaps_log(1,"plugin_c_pep_optequal: invalid invocation with one or two NULL pointers");
		return -1;
	}

	n = strlen(s);
	m = strlen(t);
	
	/* it should be an option and start with at least one dash */
	if ('-' != *p ) return 0;

	/* eat at most two dashes from p */
	if ('-' == *p) p++;
	if ('-' == *p) p++;

	n = strlen(p);
	if ( n != m ) return 0;

        if ( strncasecmp(p, t, m) == 0 ) return 1;
	return 0;
}


/******************************************************************************
Function:   lcmaps_plugin_c_pep_cleanse_pemstring(char *in, char **path)
Description:
    Remove any data but certs from a PEM string and store the result in
    a newly-alloced buffer. Return a pointer to the buffer as well, or
    NULL on failure. In case of failure, no malloc will persist.
******************************************************************************/

char *lcmaps_plugin_c_pep_cleanse_pemstring(char *in, char **destp)
{
    char *s,*t,*p;
    int  i;
    static char *pem_begin_certificate = "-----BEGIN CERTIFICATE-----";
    static char *pem_end_certificate = "-----END CERTIFICATE-----";

    if ( !in || !destp ) {
        return NULL;
    }
    s = malloc(strlen(in)+1);
    if ( !s ) {
      return NULL;
    }

    t = in;
    p = s;
    while ( *t ) {
        if ( '-' == *t ) {
            if ( strncmp(t,pem_begin_certificate,strlen(pem_begin_certificate)) == 0 ) {
                /* a cert blob is starting here, copy until the end of the cert */
                for( i = 0; *t && (i < strlen(pem_begin_certificate)); i++ ) *(p++) = *(t++);
                while ( 1 ) {
                    if ( ! *t ) break;
                    if ( '-' == *t && ( strncmp(t,pem_end_certificate,strlen(pem_end_certificate)) == 0 ) ) {
                        for( i = 0; *t && (i < strlen(pem_end_certificate)); i++ ) *(p++) = *(t++);
                        /* include any training new line, usually one has at least one, then get out */
                        while ( '\r' == *t || '\n' == *t ) *(p++) = *(t++);
                        break;
                    }
                    *(p++) = *(t++);
                }
            } else 
                t++;
        } else
            t++;
    }
    *p = '\0';

    *destp = s;
    return s;
}


/******************************************************************************
Function:   plugin_c_pep_set_capath_default(char **path)
Description:
    Determine the default CApath directory, if there is any. It will look
    for X509_CERT_DIR, then for /etc/grid-security/certificates
Parameters:
    address of the pointer to string to be filled with the directory name
Returns:
    the name of the directory, pointing to the piece of memory alloc'ed by
    this function
    NULL if no default directory was found (no memory will have been alloced and
    *path remains unmodified)
******************************************************************************/

char *plugin_c_pep_set_capath_default(char **p)
{
    char  *dirname;
    int   len, rc;
    struct stat sbuf;
    char  *targets[3];
    int   i;
    char  tmpname[PATH_MAX];

    if ( !p ) {
        lcmaps_log(1,"plugin_c_pep_set_capath_default: invalid invocation: address to a pointer to buffer is nil");
        return NULL;
    }

    /* build the search path list, ensure that X509_CERT_DIR is at the end
       since getenv(3) semantics define the output buffer to be volatile,
       i.e. a direct result from getenv(3) may only be in the list once */
    targets[2] = "/etc/grid-security/certificates";
    if ( getenv("HOME") && ( strlen(getenv("HOME")) < (PATH_MAX-21)) ) {
        sprintf(tmpname,"%s/.globus/certificates",getenv("HOME"));
        targets[1] = tmpname;
    } else
        targets[1] = NULL;
    targets[0] = getenv("X509_CERT_DIR");

    for ( i = 0 ; i < 3 ; i++ ) {
        if ( targets[i] ) {
            len = strlen(targets[i]);
            rc = stat(targets[i],&sbuf);
            if ( rc ) { /* does not exist or any other error, dont use this */
                continue;
            }
            if ( !S_ISDIR(sbuf.st_mode) ) {
                lcmaps_log(1,"plugin_c_pep_set_capath_default: %s exists but is not a directory, ignored",targets[i]);
                continue;
            }
            len = strlen(targets[i]);
            dirname = malloc(len+1);
            if ( !dirname ) {
                lcmaps_log(1,"plugin_c_pep_set_capath_default: out of memory when mallocing space for copy of default CA dir %s",targets[i]);
                *p = NULL;
                return NULL;
            }
            strncpy(dirname,targets[i],len);
            dirname[len+1] = '\0';
            *p = dirname;
            return dirname;
        }
    }
   
    return NULL;
}


char * readProxyFile (const char * proxyfile)
{
    int                 fd = -1;
    char *              pem_buf = NULL;
    int                 nbytes = 0;
    int                 rc = 0;
    struct stat         s1;
    struct stat         s2;

    uid_t stored_real_uid = -1;
    uid_t stored_eff_uid  = -1;


    /* Downgrade effective privileges to cope with NFS mounted file systems with root squashing */
    downgradeEffectiveToRealUid (&stored_real_uid, &stored_eff_uid);


    /* Read PEM string */
    if ((fd = open (proxyfile, O_RDONLY)) < 0)
    {
        lcmaps_log (2, "%s: Could not open file for reading \"%s\". Reason: %s\n", __func__, proxyfile, strerror(errno));
        return NULL;
    }
    else
    {
        /* Stat the proxy file */
        if (fstat (fd, &s1) < 0)
        {
            lcmaps_log (2, "%s: Could not stat opened file. Reason: %s\n", __func__, proxyfile, strerror(errno));
            pem_buf = NULL;
        }
        else
        {
            /* Allocate buffer with the size of the opened file */
            pem_buf = (char *) calloc (1, sizeof (char) * s1.st_size + 1);

            /* Read into the buffer */
            while (1)
            {
                if ((rc = read(fd, &pem_buf[nbytes], s1.st_size - nbytes)) < 0)
                {
                    lcmaps_log (2, "%s: Read error on file \"%s\". Reason: %s.", __func__, proxyfile, strerror(errno));
                    free (pem_buf);
                    pem_buf = NULL;
                    break;
                }
                else if (rc > 0)
                {
                    /* Let's try to read some more */
                    nbytes += rc;
                    continue;
                }
                else
                {
                    /* Done reading. rc should be zero, as bigger and smaller is covered */
                    break;
                }
            }

            /* Perform safety check to notice tampering with the file */
            if (pem_buf)
            {
                /* Stat the proxy file */
                if (fstat (fd, &s2) < 0)
                {
                    lcmaps_log (2, "%s: Could not re-stat opened file \"%s\". Reason: %s\n", __func__, proxyfile, strerror(errno));
                    free (pem_buf);
                    pem_buf = NULL;
                }
                else
                {
                    /* Safety check, simple check to see if the file has not been tampered with during the read. */
                    if ((s1.st_dev == s2.st_dev) &&
                        (s1.st_ino == s2.st_ino) &&
                        (s1.st_size == s2.st_size))
                    {
                        lcmaps_log_debug (2, "%s: File \"%s\" was read safely.", __func__, proxyfile);
                    }
                    else
                    {
                        lcmaps_log_debug (5, "%s: File \"%s\" has been moved or written to during the read.\n", __func__, proxyfile);
                        free (pem_buf);
                        pem_buf = NULL;
                    }
                }
            }
        }

        /* Close the file */
        close(fd);
    }

    /* Restore privileges to previous state */
    upgradeEffectiveToRealUid (stored_real_uid, stored_eff_uid);

    /* Return the buffer (or NULL) */
    return pem_buf;
}


/*  When the proxy is located on an NFS mount and on the server side the root squash
 *  option has been enabled, the effective uid is mapped to user 'nobody' which should
 *  not be able to read the proxy file. To work around this problem, the effective
 *  uid of the process is changed to that of the calling user and once glexec is done,
 *  the saved uid is used to restore the identity of the process,
 */ 
int downgradeEffectiveToRealUid (uid_t * real_uid, uid_t * saved_uid)
{
    *real_uid = getuid();
    if (*real_uid != 0)
    {
        /* Save it */
        *saved_uid = geteuid();
        if (seteuid(*real_uid))
        {
            lcmaps_log(1, "downgradeEffectiveToRealUid: Error on downsizing with seteuid()\n");
            return 1;
        }
    }
    return 0;
}

int upgradeEffectiveToRealUid (uid_t real_uid, uid_t saved_uid)
{
    /*  Do not forget to put back the original effective uid on the process. */
    if (real_uid != 0) 
    {
        if (seteuid(saved_uid))
        {
            lcmaps_log(1, "upgradeEffectiveToRealUid: Error on returning seteuid()\n");
            return 1;
        }
    }
    return 0;
}

